/*----------------------------------------------------------------------
  File    : kelr.c
  Contents: kernel estimation for logistic regression
  Author  : Christian Borgelt
  History : 2003.10.29 file created from file mcli.c
            2004.04.22 special mode for distance listing added
            2007.02.13 adapted to modified module tabread
            2011.02.02 adapted to modified utility modules
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#include "params.h"
#include "radfn.h"
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "kelr"
#define DESCRIPTION "kernel estimation for logistic regression"
#define VERSION     "version 2.1 (2014.10.24)         " \
                    "(c) 2003-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARAM     (-9)        /* invalid parameter value */
#define E_PATCNT   (-10)        /* pattern file is empty */
#define E_VALCNT   (-11)        /* no input variables */
#define E_CLASS    (-12)        /* too many classes */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARAM    -9 */  "invalid parameter value %g",
  /* E_PATCNT  -10 */  "pattern file is empty",
  /* E_VALCNT  -11 */  "no input variables",
  /* E_CLASS   -12 */  "too many classes",
  /*    -13 to -15 */  NULL, NULL, NULL,
  /* E_VALUE   -16 */  "%s:%d(%d): invalid value '%s'",
  /* E_FLDCNT  -17 */  "%s:%d(%d): wrong number of fields/columns",
  /* E_RECCNT  -18 */  "%s:%d(%d): wrong number of records/rows",
  /*           -19 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static TABREAD *tread  = NULL;  /* table reader */
static MATRIX  *mat    = NULL;  /* matrix of training patterns */
static double  *kes[2] = {NULL, NULL}; /* arrays of kernel estimates */
static FILE    *out    = NULL;  /* output file */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (kes[0])free(kes[0]);         \
  if (mat)   mat_delete(mat);      \
  if (tread) trd_delete(tread, 1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, buffers */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *blanks  = NULL;      /* blanks */
  CCHAR   *fldseps = NULL;      /* field  separators */
  CCHAR   *recseps = NULL;      /* record separators */
  CCHAR   *comment = NULL;      /* comment characters */
  char    seps[4]  = "  \n";    /* separators for output */
  double  params[] = {2, 0};    /* kernel function parameters */
  double  radius   = 1;         /* radius of kernel function */
  double  thresh   = 0;         /* threshold for class assignment */
  int     digs     = 6;         /* number of significant digits */
  int     invert   = 0;         /* flag for threshold inversion */
  int     dist     = 0;         /* flag for distance listing */
  DIMID   patcnt   = 0;         /* number of patterns */
  DIMID   valcnt   = 0, n;      /* number of values per pattern */
  RADFN   *radfn   = rf_gauss;  /* radial kernel function */
  double  *v1, *v2;             /* vectors/data points to compare */
  DIMID   c1 = 0, c2, cnt;      /* classes of vectors, class counter */
  double  d, e;                 /* for distance computation */
  clock_t t;                    /* for time measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] infile outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c       use Cauchy kernel                 "
                    "(default: Gaussian)\n");
    printf("-p#:#    kernel function parameters        "
                    "(default: %g:%g)\n", params[0], params[1]);
    printf("-s       size/radius of kernel function    "
                    "(default: %g)\n", radius);
    printf("-t#      threshold for class assignment    "
                    "(default: > %g)\n", thresh);
    printf("-i       invert threshold (<= instead of >)\n");
    printf("-d       special mode: list all distances\n");
    printf("-o#      number of significant digits      "
                    "(default: %d)\n", digs);
    printf("-r#      record  separators                "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                "
                    "(default: \"#\")\n");
    printf("infile   file to read (no header, only numbers)\n");
    printf("outfile  file to write kernel estimates to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'c': radfn  = rf_cauchy;             break;
          case 'p': getdblvec(s, &s, 3, params);    break;
          case 's': radius = strtod(s, &s);         break;
          case 't': thresh = strtod(s, &s);         break;
          case 'i': invert = 1;                     break;
          case 'd': dist   = 1;                     break;
          case 'o': digs   = (int)strtol(s, &s, 0); break;
          case 'b': optarg = &blanks;               break;
          case 'f': optarg = &fldseps;              break;
          case 'r': optarg = &recseps;              break;
          case 'C': optarg = &comment;              break;
          default : error(E_OPTION, *--s);          break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_inp = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (k != 2) error(E_ARGCNT);  /* and the number of arguments */
  if (params[0] <= 0) error(E_PARAM, params[0]);
  if (params[1] <  0) error(E_PARAM, params[1]);
  if (radius    <= 0) error(E_PARAM, radius);
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read training patterns --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* set the separator characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  k = mat_readx(&mat, tread, -1, -1);    /* read the matrix */
  if (k) error(k, TRD_INFO(tread));
  trd_delete(tread, 1);         /* close the input file and */
  tread  = NULL;                /* delete the table reader */
  patcnt = mat_rowcnt(mat);     /* get the number of data points */
  valcnt = mat_colcnt(mat);     /* and their dimensionality */
  if (patcnt <= 0) error(E_PATCNT);
  if (valcnt <= 1) error(E_VALCNT);
  fprintf(stderr, "[%"DIMID_FMT" pattern(s)] done", patcnt);
  fprintf(stderr, " [%.2fs].\n", SEC_SINCE(t));

  /* --- open the output file --- */
  t = clock();                  /* start the timer */
  if (fn_out && (strcmp(fn_out, "-") == 0)) fn_out = "";
  if (fn_out && *fn_out) { out = fopen(fn_out, "wb"); }
  else                   { out = stdout; fn_out = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_out);
  if (!out) error(E_FOPEN, fn_out);

  /* --- compute kernel estimates --- */
  if (dist)                     /* open the output file */
    fputs("dist class\n", out); /* print the table header */
  else {                        /* if kernel estimation */
    kes[0] = (double*)calloc((size_t)(patcnt+patcnt), sizeof(double));
    if (!kes[0]) error(E_NOMEM);
    kes[1] = kes[0] +patcnt;    /* create kernel estimate arrays */
  }                             /* and organize the memory */
  radius *= radius;             /* compute variance and */
  if (radius <= 0) radius = 1;  /* check and adapt it */
  for (cnt = 0, i = 0; i < patcnt; i++) {
    v1 = mat_row(mat, i);       /* traverse the matrix rows */
    if (!dist) {                /* if kernel estimation */
      c1 = ((v1[valcnt-1] > thresh) ? 1 : 0) ^ invert;
      if (c1 <= 0) cnt++;       /* count vectors of class 1 */
      kes[c1][i] += radfn(0, params);
    }                           /* count vector for itself */
    for (k = i; --k >= 0; ) {   /* traverse preceding matrix rows */
      v2 = mat_row(mat, k);     /* (consider each pair exactly once) */
      for (d = 0, n = valcnt-1; --n >= 0; ) {
        e = v1[n] -v2[n]; d += e*e; }
      if (dist) {               /* if distance computation */
        c1 = (1 << (int)v1[valcnt-1]) | (1 << (int)v2[valcnt-1]);
        c2 = c1 -('z'-'a'+1);   /* compute class indicator */
        if      (c2 <  0)         c1 = 'a' -1 +c1;
        else if (c2 <= 'Z'-'A'+1) c1 = 'A' -1 +c2;
        else error(E_CLASS);    /* code the class indicator */
        fprintf(out, "%g %c\n", sqrt(d), (int)c1); }
      else {                    /* if kernel estimation */
        c2 = ((v2[valcnt-1] > thresh) ? 1 : 0) ^ invert;
        kes[c2][i] += e = radfn(d /radius, params);
        kes[c1][k] += e;        /* compute the distance between the */
      }                         /* two patterns and sum the values */
    }                           /* of the kernel function */
  }
  if (!dist) {                  /* if kernel estimation */
    for (n = patcnt; --n >= 0; ) { /* traverse the patterns */
      v1 = mat_row(mat, n);     /* get the next pattern and */
      e = kes[0][n] +kes[1][n]; /* sum the kernel function values */
      v1[valcnt-1] = (e > 0) ? kes[1][n] /e : 1;
    }                           /* compute the probability estimate */
  }                             /* from the aggregated weights */

  /* --- write kernel estimates --- */
  if (!dist)                    /* if to compute kernel estimation, */
    mat_write(mat, out, digs, seps);   /* write the modified matrix */
  if (((out == stdout) ? fflush(out) : fclose(out)) != 0)
    error(E_FWRITE, fn_out);    /* close the output file */
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
