/*----------------------------------------------------------------------
  File    : regress.c
  Contents: multivariate polynomial regression management
  Author  : Christian Borgelt
  History : 2001.10.13 file created
            2001.10.15 first version completed
            2001.10.19 bug in function reg_desc() removed
            2001.10.20 computation of the sum of squared errors added
            2001.10.26 function reg_sse() added, reg_calc() modified
            2007.04.04 extended regression functions added
            2007.04.12 functions reg_parse() and reg_parsex() added
            2007.05.30 linear regression treated separately in prods()
            2009.11.19 function reg_showles() added for debugging
            2013.05.06 static and dynamic Tikhonov regularization added
            2013.08.13 adapted to definition of type DIMID in matrix.h
            2015.07.30 logistic regression with maximum likelihood added
            2015.07.31 functions reg_minmax() and reg_reglim() added
            2015.08.08 bug in function reg_exec() fixed (logistic fn.)
            2015.09.18 bug in regularization loop fixed (reg_solve())
            2015.11.30 nominal values mapped with 1-in-(n-1) encoding
            2018.09.13 extended functions (attset, attmap) removed
            2018.10.12 gradient methods and N-R for error function added
----------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "regress.h"

#ifdef _MSC_VER
#ifndef strtoll
#define strtoll     _strtoi64   /* string to long long int */
#endif
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define REG_MAX     ((DIMID)sqrt((double)(DIMID_MAX/4)/sizeof(double)))

#define logistic(x) (((x) > -709.0) ? 1.0 /(1.0 +exp(-(x))) : 0.0)

/* --- error codes --- */
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_INVDIM    (-17)       /* invalid number of dimensions */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef REG_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_INVDIM  -17 */  "#invalid number of dimensions"
};
#endif

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static DIMID count (REGRESS *reg)
{                               /* --- compute number of coefficients */
  int   i;                      /* loop variable for degree */
  DIMID k;                      /* loop variable for dimensions */

  assert(reg && reg->tab);      /* check the function arguments */
  if (reg->dim <= 1) return 1;  /* if there is no regressor */
  if ((reg->dim <= 2)           /* if single variable regression */
  ||  (reg->deg <= 2)) {        /* or linear/quadratic regression */
    if (reg->deg <= 1) return reg->dim;
    if (reg->dim <= 2) return reg->deg +1;
    return (reg->dim *(reg->dim +1)) >> 1;
  }                             /* compute the number directly */
  for (k = reg->dim; --k >= 0;) /* initialize table element n-1 to */
    reg->tab[k] = k+1;          /* n choose 1 = n (sizes for deg = 1) */
  for (i = reg->deg-1; --i >= 0; ) {  /* compute the number as */
    for (k = 0; ++k < reg->dim; ) {   /* (dim +deg -1) choose deg */
      if (reg->tab[k] > DIMID_MAX -reg->tab[k-1])
        return -1;              /* check the number of coefficients */
      reg->tab[k] += reg->tab[k-1];
    }                           /* compute next binomial coefficient */
  }                             /* by dynamic programming */
  k = reg->tab[reg->dim-1];     /* finally get the number of params. */
  return (k > REG_MAX) ? -1 : k;/* from the last table field */
}  /* count() */

/*--------------------------------------------------------------------*/

static void prods (REGRESS *reg, double *vec)
{                               /* --- compute products of variables */
  DIMID  i, k, m;               /* loop variables */
  DIMID  *s;                    /* table of section sizes */
  double *b, *p;                /* to traverse the buffer */

  assert(reg);                  /* check the function argument */
  b = reg->buf +reg->cnt -1;    /* traverse the buffer */
  if (reg->deg <= 1) {          /* if linear regresseion */
    for (i = reg->dim-1; --i >= 0; )
      *--b = vec[i];            /* simply copy the data vector */
    return;                     /* to the internal buffer and */
  }                             /* then abort the function */
  for (s = reg->tab +(i = reg->dim-1); --i >= 0; ) {
    *--b = vec[i];              /* copy the linear terms to the end */
    *--s = i+1;                 /* and set the initial section sizes */
  }                             /* (elements to combine per variable) */
  for (k = 2; 1; ) {            /* traverse the exponent sums */
    p = b;                      /* note the previous section */
    for (i = reg->dim-1; --i >= 0; )
      for (m = s[i]; --m >= 0;) /* compute next section (next exp.) */
        *--b = p[m] *vec[i];    /* from the previous one */
    if (++k > reg->deg) break;  /* if highest degree reached, abort */
    for (i = 0; ++i < reg->dim; )
      s[i] += s[i-1];           /* compute next section sizes */
  }                             /* (dynamic programming) */
}  /* prods() */

/*--------------------------------------------------------------------*/

static int solve (REGRESS *reg, double *res)
{                               /* --- solve equation system */
  DIMID  r;                     /* loop variables for matrix rows */
  double x, y;                  /* regularization value, limit */

  assert(reg);                  /* check the function arguments */
  y = reg->lim *reg->lim;       /* compute regularization limit and */
  x = reg->reg *reg->reg;       /* get initial regularization value */
  mat_diaget(reg->mat,reg->vec);/* save diagonal for regularization */
  if (reg->reg > 0) {           /* if to use Tikhonov regularization */
    for (r = 0; r < reg->dim; r++) mat_inc(reg->mat, r, r, x);
    x *= 4;                     /* apply initial regularization */
  }                             /* and update regularization value */
  while (1) {                   /* Hessian matrix decomposition loop */
    r = mat_chdecom(reg->mat, reg->mat);
    if (r == 0) break;          /* try Cholesky decomposition */
    if ((reg->reg >= 0) || (x > y)) return -1;
    mat_diaset(reg->mat, reg->vec); /* on failure restore diagonal */
    for (r = 0; r < reg->dim; r++) mat_inc(reg->mat, r, r, x);
    x *= 4;                     /* add regularization to diagonal */
  }                             /* and update regularization value */
  mat_chsubst(res, reg->mat, reg->rhs);
  return 0;                     /* return 'ok' */
}  /* solve() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

REGRESS* reg_create (DIMID dim, int deg)
{                               /* --- create a regression object */
  REGRESS *reg;                 /* created regression object */

  assert((dim > 0) && (deg >= 0));  /* check the function arguments */
  reg = (REGRESS*)malloc(sizeof(REGRESS));
  if (!reg) return NULL;        /* create the base structure */
  reg->dim = (deg > 0) ? dim : 1;   /* note the number of variables */
  reg->deg = deg;               /* and the degree of the polynomial */
  reg->sse = -1;                /* invalidate sum of squared errors */
  reg->min = reg->max = 0;      /* and logit transformation */
  reg->reg = reg->lim = 0;      /* default: no regularization */
  reg->tab = (DIMID*)malloc((size_t)dim *sizeof(DIMID));
  if (!reg->tab)    {                 free(reg); return NULL; }
  reg->cnt = count(reg);        /* compute the number of parameters */
  if (reg->cnt < 0) { free(reg->tab); free(reg); return NULL; }
  reg->vec = (double*)malloc((size_t)(3*reg->cnt +dim) *sizeof(double));
  if (!reg->vec)    { free(reg->tab); free(reg); return NULL; }
  reg->buf = reg->vec +dim;     /* organize the memory into buffers */
  reg->rhs = reg->buf +reg->cnt;/* for data and power products */
  reg->cfs = reg->rhs +reg->cnt;/* and a right hand side vector */
  reg->mat = mat_create(reg->cnt, reg->cnt);
  if (!reg->mat) {              /* create the regression matrix */
    free(reg->vec); free(reg->tab); free(reg); return NULL; }
  return reg;                   /* return the created reg. object */
}  /* reg_create() */

/*--------------------------------------------------------------------*/

void reg_delete (REGRESS *reg)
{                               /* --- delete a regression object */
  assert(reg);                  /* check the function argument */
  mat_delete(reg->mat);         /* delete the regression matrix, */
  free(reg->vec);               /* the internal buffers, */
  free(reg->tab);               /* the table of sizes, */
  free(reg);                    /* and the base structure */
}  /* reg_delete() */

/*--------------------------------------------------------------------*/

void reg_minmax (REGRESS *reg, double min, double max)
{                               /* --- set span of (logit) transform */
  assert(reg);                  /* check the function arguments */
  reg->min = min; reg->max = max;
}  /* reg_minmax() */

/*--------------------------------------------------------------------*/

void reg_reglim (REGRESS *reg, double val, double lim)
{                               /* --- set regularization parameters */
  assert(reg);                  /* check the function arguments */
  reg->reg = val; reg->lim = lim;
}  /* reg_reglim() */

/*--------------------------------------------------------------------*/

void reg_init (REGRESS *reg)
{                               /* --- (re)initialize regression */
  mat_init(reg->mat, MAT_ZERO|MAT_VECTOR|MAT_WEIGHT, NULL);
  reg->sse = -1;                /* clear matrix and sum of errors */
}  /* reg_init() */

/*--------------------------------------------------------------------*/

int reg_aggvec (REGRESS *reg, double *vec, double wgt)
{                               /* --- aggregate a data vector */
  DIMID  i, k;                  /* loop variables */
  double *b;                    /* to traverse the buffer */
  double t, o = 0;              /* temporary buffers */

  assert(reg && vec && (wgt >= 0)); /* check the function arguments */
  if (reg->max > reg->min) {    /* if logit transformation, */
    b = vec +reg->dim-1; o = *b;/* get the output value */
    if ((o <= reg->min) || (o >= reg->max))
      return -1;                /* check whether in range */
    *b = log((o -reg->min) / (reg->max -o));
  }                             /* transform the output value */
  if      (reg->deg <= 1)       /* --- if linear regression */
    mat_addmpx(reg->mat, vec, wgt);  /* aggregate the data vector */
  else if (reg->dim <= 2) {     /* --- if single variable regression */
    b  = reg->buf +(i = reg->deg);   /* compute the needed powers */
    *b = vec[1]; *--b = t = vec[0];  /* of the regressor variable */
    while (--i > 0) { *--b = t *= vec[0]; }
    mat_addmpx(reg->mat, b, wgt); }  /* aggregate vector of powers */
  else if (reg->deg <= 2) {     /* --- if quadratic regression */
    b = reg->buf +reg->cnt;     /* traverse the buffer */
    for (i = reg->dim; --i >= 0; )
      *--b = vec[i];            /* copy linear terms to the end */
    for (i = reg->dim-1; --i >= 0; )
      for (k = i+1; --k >= 0; ) /* compute squared values */
        *--b = vec[i]*vec[k];   /* and mixed terms */
    mat_addmpx(reg->mat, b, wgt); }  /* aggregate vector of products */
  else {                        /* --- if gen. polynomial regression */
    prods(reg, vec);            /* compute products of the variables */
    reg->buf[reg->cnt-1] = vec[reg->dim-1];  /* and store the output */
    mat_addmpx(reg->mat, reg->buf, wgt);
  }                             /* aggregate the vector of products */
  if (reg->max > reg->min)      /* if logit transformation, */
    vec[reg->dim-1] = o;        /* restore the output value */
  return 0;                     /* return 'ok' */
}  /* reg_aggvec() */

/*--------------------------------------------------------------------*/

DIMID reg_aggmat (REGRESS *reg, MATRIX *data)
{                               /* --- aggregate a data matrix */
  DIMID i;                      /* loop variable */

  assert(reg && data);          /* check the function arguments */
  for (i = mat_rowcnt(data); --i >= 0; )
    if (reg_aggvec(reg, mat_row(data, i), mat_rgetwgt(data, i)) != 0)
      return i;                /* aggregate all matrix rows */
  return 0;                    /* return 'ok' */
}  /* reg_aggmat() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void reg_showles (REGRESS *reg)
{                               /* --- show linear equation system */
  DIMID r, c;                   /* loop variables (row and column) */

  for (r = 0; r < reg->dim; r++) {  /* traverse the rows */
    for (c = 0; c < reg->dim; c++)  /* traverse the columns */
      printf("%8.4f ", mat_get(reg->mat, r, c));
    printf("| %8.4f\n", reg->rhs[r]);
  }                             /* print the linear equation system */
}  /* reg_showles() */

#endif
/*--------------------------------------------------------------------*/

int reg_solve (REGRESS *reg)
{                               /* --- solve regression eq. system */
  assert(reg);                  /* check the function argument */
  reg->so2 = mat_regress(reg->mat, reg->rhs, reg->mat);
  return solve(reg, reg->cfs);  /* set up and solve equation system */
}  /* reg_solve() */

/*--------------------------------------------------------------------*/

double reg_sse (REGRESS *reg)
{                               /* --- compute sum of squared errors */
  DIMID  i, k;                  /* loop variables */
  double sse, t = 0;            /* sum of squared errors, buffer */
  double *r, *c;                /* to access rhs and coefficients */
  double *p;                    /* to access the matrix rows */

  assert(reg);                  /* check the function argument */
  if (reg->sse >= 0)            /* if sum of squared errors is known, */
    return reg->sse;            /* simply return it */
  r = reg->rhs; c = reg->cfs;   /* compute the sum of squared errors */
  i = reg->cnt -1;              /* from the aggregated vectors */
  sse = reg->so2 +c[i] *c[i] *mat_getwgt(reg->mat) -c[i] *r[i] *2;
  while (--i >= 0) {            /* traverse the matrix rows */
    p = mat_row(reg->mat, i);   /* except the last one */
    for (t = 0, k = i+1; --k >= 0; )
      t += p[k] *p[k];          /* recompute the matrix diagonal */
    sse += c[i] *c[i] *t;       /* from the Cholesky decomposition */
    for (t = -r[i], k = reg->cnt; --k > i; )
      t += c[k] *p[k];          /* compute c^T *M *c -2 *c^T *r +y^2, */
    sse += c[i] *t *2;          /* where M is the original matrix and */
  }                             /* r the right hand side of the LES */
  reg->sse = (sse < 0) ? 0 : sse;
  return reg->sse;              /* return the sum of squared errors */
}  /* reg_sse() */

/*--------------------------------------------------------------------*/

DIMID reg_logopt (REGRESS *reg, MATRIX *data, int mode,
                  double lrate, double limit, int maxiter)
{                               /* --- optimized logistic regression */
  int    k;                     /* loop variable for updates */
  DIMID  r, c;                  /* loop variables for rows/columns */
  DIMID  i, n;                  /* loop variables for tuples */
  double *vec;                  /* to traverse the data points */
  double trg, out;              /* actual and computed target value */
  double err;                   /* error of output value */
  double x;                     /* buffer for intermediate values */
  double wgt;                   /* tuple weight (multiplicity) */
  double old;                   /* previous sum of squared errors */

  assert(reg && data);          /* check the function arguments */
  reg_init(reg);                /* initialize the regression object */
  r = reg_aggmat(reg, data);    /* aggregate the data matrix */
  if (r) return r;              /* and check for an error */
  if (reg_solve(reg) != 0)      /* solve the linear equation system */
    return -1;                  /* and check for an error */
  if (mode == REG_NONE)         /* if no subsequent optimization, */
    return 0;                   /* simply abort the function */
  old = reg->sse;               /* note the old sum of errors */
  if (reg->max <= reg->min) {   /* set default logistic transform */
    reg->min = 0; reg->max = 1; }
  n = mat_rowcnt(data);         /* get the number of data points */

  /* --- gradient descent on error function --- */
  if     ((mode == REG_ERR_GD)  /* if one of the gradient methods */
  ||      (mode == REG_LLH_GA)) {
    for (k = 0; k < maxiter; k++) { /* compute maxiter updates */
      reg->sse = 0.0;           /* init. sum of squared errors */
      vec_init(reg->rhs, reg->dim, -1);  /* init. the gradient */
      for (i = 0; i < n; i++) { /* traverse the matrix rows */
        vec = mat_row(data, i); /* get the next data point */
        wgt = mat_rgetwgt(data, i); /* get weight/multiplicity */
        trg = vec[r = reg->dim-1];  /* an the target value */
        out = reg->cfs[r];      /* compute the linear function */
        while (--r >= 0) out += reg->cfs[r] *vec[r];
        trg = (trg     -reg->min)   /* transform target value */
            / (reg->max-reg->min);  /* to unit interval [0,1] */
        out = logistic(out);    /* compute output value and */
        err = (x = trg-out) *wgt;          /* and its error */
        reg->sse += x *err;     /* update sum of squared errors */
        if (mode == REG_ERR_GD) /* include derivative of logistic */
          err *= out *(1-out);  /* function for gradient descent */
        reg->rhs[r = reg->dim-1] += err;
        while (--r >= 0) reg->rhs[r] += vec[r] *err;
      }                         /* update the gradient */
      vec_add(reg->cfs, reg->dim, reg->cfs, lrate, reg->rhs);
      if (fabs(reg->sse -old) < limit)
        break;                  /* compute and apply changes */
      old = reg->sse;           /* and check change of error sum */
    }                           /* against the given threshold */
  }  /* if (mode == REG_ERR_GD) .. */

  /* --- Newton-Raphson on error function --- */
  else if (mode == REG_ERR_NR) {
    for (k = 0; k < maxiter; k++) { /* compute maxiter updates */
      reg->sse = 0.0;           /* init. sum of squared errors */
      vec_init(reg->rhs, reg->dim, -1);   /* init. the gradient */
      mat_init(reg->mat, MAT_ZERO, NULL); /* and the Hessian matrix */
      for (i = 0; i < n; i++) { /* traverse the matrix rows */
        vec = mat_row(data, i); /* get the next data point */
        wgt = mat_rgetwgt(data, i); /* get weight/multiplicity */
        trg = vec[r = reg->dim-1];  /* an the target value */
        out = reg->cfs[r];      /* compute the linear function */
        while (--r >= 0) out += reg->cfs[r] *vec[r];
        trg = (trg     -reg->min)   /* transform target value */
            / (reg->max-reg->min);  /* to unit interval [0,1] */
        out = logistic(out);    /* compute output value and */
        err = (x = trg-out) *wgt;          /* and its error */
        reg->sse += x *err;     /* update sum of squared errors */
        x   = out *(1-out);     /* logistic function derivative */
        err = -2 *x *err;       /* get factor for gradient */
        x   = -2 *(trg -2 *(trg+1) *out +3 *out *out) *x *wgt;
        reg->rhs[r = reg->dim-1] += err;
        mat_inc(reg->mat, r, r, x);
        while (--r >= 0) {      /* traverse the data values */
          reg->rhs[r] += vec[r] *err;
          mat_inc(reg->mat, r, reg->dim-1, vec[r] *x);
          for (c = reg->dim-1; --c >= r; )
            mat_inc(reg->mat, r, c, vec[r] *vec[c] *x);
        }                       /* update the gradient */
      }                         /* and the Hessian matrix */
      if (solve(reg, reg->vec) != 0)
        return -1;              /* solve the equation system */
      vec_add(reg->cfs, reg->dim, reg->cfs, -1.0, reg->vec);
      if (fabs(reg->sse -old) < limit)
        break;                  /* compute and apply changes */
      old = reg->sse;           /* and check change of error sum */
    }                           /* against the given threshold */
  }  /* if (mode == REG_ERR_NR) .. */

  /* --- Newton-Raphson on likelihood function --- */
  else /* (mode == REG_LLH_NR) */ {
    for (k = 0; k < maxiter; k++) { /* compute maxiter updates */
      reg->sse = 0.0;           /* init. sum of squared errors */
      vec_init(reg->rhs, reg->dim, -1);   /* init. the gradient */
      mat_init(reg->mat, MAT_ZERO, NULL); /* and the Hessian matrix */
      for (i = 0; i < n; i++) { /* traverse the matrix rows */
        vec = mat_row(data, i); /* get the next data point */
        wgt = mat_rgetwgt(data, i); /* get weight/multiplicity */
        trg = vec[r = reg->dim-1];  /* an the target value */
        out = reg->cfs[r];      /* compute the linear function */
        while (--r >= 0) out += reg->cfs[r] *vec[r];
        trg = (trg     -reg->min)   /* transform target value */
            / (reg->max-reg->min);  /* to unit interval [0,1] */
        out = logistic(out);    /* compute output value and */
        err = (x = trg-out) *wgt;          /* and its error */
        reg->sse += x *err;     /* update sum of squared errors */
        x   = out *(1-out) *wgt;/* get factor for Hessian matrix */
        reg->rhs[r = reg->dim-1] += err;
        mat_inc(reg->mat, r, r, x);
        while (--r >= 0) {      /* traverse the data values */
          reg->rhs[r] += vec[r] *err;
          mat_inc(reg->mat, r, reg->dim-1, vec[r] *x);
          for (c = reg->dim-1; --c >= r; )
            mat_inc(reg->mat, r, c, vec[r] *vec[c] *x);
        }                       /* update the gradient */
      }                         /* and the Hessian matrix */
      if (solve(reg, reg->vec) != 0)
        return -1;              /* solve the equation system */
      vec_add(reg->cfs, reg->dim, reg->cfs, 1.0, reg->vec);
      if (fabs(reg->sse -old) < limit)
        break;                  /* compute and apply changes */
      old = reg->sse;           /* and check change of error sum */
    }                           /* against the given threshold */
  }  /* if (mode == REG_LLH_NR) .. */
  return 0;                     /* return 'ok' */
}  /* reg_logopt() */

/*--------------------------------------------------------------------*/

double reg_exec (REGRESS *reg, double *vec)
{                               /* --- execute a regression function */
  DIMID  i, k;                  /* loop variables */
  double *b, *c;                /* to traverse the buffer */
  double r, t;                  /* value of regression func., buffer */

  assert(reg && vec);           /* check the function arguments */
  if      (reg->deg <= 1) {     /* --- if linear regression */
    r = reg->cfs[i = reg->dim-1];  /* compute linear function */
    while (--i >= 0) r += reg->cfs[i] *vec[i]; }
  else if (reg->dim <= 2) {     /* --- if single variable regression */
    c = reg->cfs +(i = reg->deg);
    r = *c;                     /* multiply the different powers */
    r += *--c *(t = vec[0]);    /* with the computed coefficients */
    while (--i >  0) r += *--c *(t *= vec[0]); }
  else if (reg->deg <= 2) {     /* --- if quadratic regression */
    c = reg->cfs +reg->cnt -1;  /* traverse the coefficients */
    r = *c;                     /* start with the constant term */
    for (i = reg->dim-1; --i >= 0; )
      r += *--c *vec[i];        /* add the linear terms */
    for (i = reg->dim-1; --i >= 0; ) {
      for (k = i+1; --k >= 0; ) /* add the squared values */
        r += *--c *vec[i]*vec[k];  /* and the mixed terms */
    } }
  else {                        /* --- if general regression */
    prods(reg, vec);            /* compute products of variables */
    b = reg->buf +reg->cnt -1;  /* traverse the buffer */
    c = reg->cfs +reg->cnt -1;  /* and the coefficients */
    r = *c;                     /* start with the constant term */
    for (i = reg->cnt-1; --i >= 0; )
      r += *--c * *--b;         /* multiply the different products */
  }                             /* with the corresponding coeffs. */
  if (reg->max > reg->min)      /* if logistic regression, transform */
      r = (reg->max -reg->min) *logistic(r) +reg->min;
  return r;                     /* return the value of the */
}  /* reg_exec() */             /* regression function */

/*--------------------------------------------------------------------*/

int reg_desc (REGRESS *reg, FILE *file, int mode, int maxlen)
{                               /* --- describe a regression function */
  DIMID i, k;                   /* loop variables */
  int   len, l;                 /* loop variables for comments */
  DIMID *e      = NULL;         /* table of exponents */
  char  *indent = "";           /* indentation string */

  assert(reg && file);          /* check the function arguments */

  /* --- print header (as a comment) --- */
  if (mode & REG_TITLE) {       /* if the title flag is set */
    len = (maxlen > 0) ? maxlen -2 : 70;
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fputs("\n  regression function\n", file);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */

  /* --- print polynomial --- */
  fputs("regression = {\n", file);
  indent = "  ";                /* print the basic parameters */
  fprintf(file, "%svars   = %"DIMID_FMT";\n", indent, reg->dim);
  fprintf(file, "%sdegree = %d;\n",           indent, reg->deg);
  if (reg->max > reg->min)      /* print maximum for logit transform. */
    fprintf(file, "%slogit  = %g, %g;\n", indent, reg->min, reg->max);
  if (mode & REG_EXPOS) {       /* if to print exponents */
    e = memset(reg->tab, 0, (size_t)(reg->dim-2) *sizeof(DIMID));
    e[reg->dim-2] = reg->deg+1; /* init. the table of exponents to */
  }                             /* x_n^(deg+1) ("previous state") */
  fprintf(file, "%scoeffs = { ", indent);
  for (i = 0; i < reg->cnt; i++) { /* list the coefficients */
    if (i > 0)         fputc(',', file);
    if (!e && (i % 3)) fputc(' ', file);
    else if (i > 0) fprintf(file, "\n%s           ", indent);
    fprintf(file, "%+.16g", reg->cfs[i]);
    if (!e) continue;           /* check whether to print exponents */
    if (e[0] > 0) {             /* update the table of exponents: */
      e[1]++; e[0]--; }         /* if first exponent is not zero, */
    else {                      /* shift a 1 from first to second */
      for (k = 0; e[++k] <= 0; );   /* find first nonzero expo. */
      e[0] = e[k]-1; e[k] = 0; e[k+1]++;
    }                           /* move exponents back to first */
    fputs(" /*", file);         /* start the list of exponents */
    for (k = 0; ++k < reg->dim; ) fprintf(file, " %"DIMID_FMT, e[k-1]);
    fputs(" */", file);         /* terminate the list of exponents */
  }
  fputs(" };\n", file);         /* terminate the coefficient list */
  fputs("};\n",  file);         /* terminate the description */
  return ferror(file) ? -1 : 0; /* return the write status */
}  /* reg_desc() */

/*--------------------------------------------------------------------*/
#ifdef REG_PARSE

static int degree (SCANNER *scan)
{                               /* --- parse degree of polynomial */
  int  deg;                     /* parsed degree of polynomial */
  char *s;                      /* end pointer for conversion */

  assert(scan);                 /* check the function argument */
  if ((scn_token(scan) != T_ID) /* check for an identifier */
  || ((strcmp(scn_value(scan), "degree") != 0)
  &&  (strcmp(scn_value(scan), "deg")    != 0)))
    SCN_ERROR(scan, E_STREXP, "degree");
  SCN_NEXT(scan);               /* consume keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_NUM (scan);               /* check for a number */
  deg = (int)strtol(scn_value(scan), &s, 0);
  if (*s || (deg < 0)) SCN_ERROR(scan, E_NUMBER);
  SCN_NEXT(scan);               /* get and consume the number */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return deg;                   /* return the parsed degree */
}  /* degree() */

/*--------------------------------------------------------------------*/

static DIMID dimen (SCANNER *scan, DIMID dim)
{                               /* --- parse the number of dimensions */
  DIMID d;                      /* parsed number */
  char  *s;                     /* end pointer for conversion */

  assert(scan);                 /* check the function argument */
  if ((scn_token(scan) != T_ID) /* check for an identifier */
  || ((strcmp(scn_value(scan), "vars") != 0)
  &&  (strcmp(scn_value(scan), "dim")  != 0)))
    SCN_ERROR(scan, E_STREXP, "vars");
  SCN_NEXT(scan);               /* consume keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_NUM (scan);               /* check for a number */
  d = strtodim(scn_value(scan), &s);
  if (*s || (d < 1))            SCN_ERROR(scan, E_NUMBER);
  if ((dim >= 0) && (dim != d)) SCN_ERROR(scan, E_INVDIM);
  dim = d;                      /* get and check the number */
  SCN_NEXT(scan);               /* consume the number */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return dim;                   /* return the parsed number */
}  /* dimen() */

/*--------------------------------------------------------------------*/

static int poly (REGRESS *reg, SCANNER *scan)
{                               /* --- parse a regression polynomial */
  DIMID i;                      /* loop variable */
  char  *s;                     /* end pointer for conversion */

  assert(reg && scan);          /* check the function arguments */
  if ((scn_token(scan) == T_ID) /* check for 'logit' */
  &&  ((strcmp(scn_value(scan), "logit") == 0)
  ||   (strcmp(scn_value(scan), "max")   == 0))) {
    SCN_NEXT(scan);             /* consume 'logit' */
    SCN_CHAR(scan, '=');        /* consume '=' */
    SCN_NUM (scan);             /* check for a number */
    reg->max = strtod(scn_value(scan), &s);
    if (*s) SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and consume logistic maximum */
    if (scn_token(scan) != ',') /* if the only value, */
      reg->min = 0;             /* set minimum to zero */
    else {                      /* if more than one value */
      reg->min = reg->max;      /* use first value as minimum */
      SCN_CHAR(scan, ',');      /* consume ',' */
      reg->max = strtod(scn_value(scan), &s);
      if (*s) SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* get and consume logistic maximum */
    }
    SCN_CHAR(scan, ';');        /* consume ';' */
  }
  if ((scn_token(scan) != T_ID) /* check for an identifier */
  || ((strcmp(scn_value(scan), "coefficients") != 0)
  &&  (strcmp(scn_value(scan), "coeffs")       != 0)))
    SCN_ERROR(scan, E_STREXP, "coeffs");
  SCN_NEXT(scan);               /* consume 'coeffs' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  for (i = 0; i < reg->cnt; i++) {  /* traverse the coefficients */
    if (i > 0) SCN_CHAR(scan, ',');
    SCN_NUM(scan);              /* check for a number */
    reg->cfs[i] = strtod(scn_value(scan), &s);
    if (*s) SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and consume */
  }                             /* a regression coefficient */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* poly() */

/*--------------------------------------------------------------------*/

REGRESS* parse (SCANNER *scan, DIMID dim)
{                               /* --- parse a regression polynomial */
  int     deg;                  /* degree of the polynomial */
  REGRESS *reg;                 /* parsed regression polynomial */

  assert(scan);                 /* check the function argument */
  scn_first(scan);              /* get the first token */
  dim = dimen(scan, dim);       /* get the number of variables */
  if (dim < 0) return NULL;
  deg = degree(scan);           /* get the degree of the polynomial */
  if (deg < 0) return NULL;
  reg = reg_create(dim, deg);   /* create a regression object */
  if (!reg)    return NULL;
  if (poly(reg, scan) != 0) {   /* parse the regression polynomial */
    reg_delete(reg); return NULL; }
  return reg;                   /* return the parsed polynomial */
}  /* reg_parse() */

/*--------------------------------------------------------------------*/

static int header (SCANNER *scan)
{                               /* --- parse header */
  int t;                        /* token */

  assert(scan);                 /* check the function arguments */
  if ((scn_token(scan) != T_ID)
  || ((strcmp(scn_value(scan), "polynomial") != 0)
  &&  (strcmp(scn_value(scan), "poly")       != 0)
  &&  (strcmp(scn_value(scan), "regression") != 0)
  &&  (strcmp(scn_value(scan), "regress")    != 0)
  &&  (strcmp(scn_value(scan), "reg")        != 0)))
    SCN_ERROR(scan, E_STREXP, "regression");
  SCN_NEXT(scan);               /* consume 'polynomial' */
  if (scn_token(scan) == '(') { /* if a perenthesis follows */
    SCN_NEXT(scan);             /* consume '(' */
    t = scn_token(scan);        /* check for an attribute name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_ATTEXP);
    SCN_NEXT(scan);             /* consume the attribute name */
    SCN_CHAR(scan, ')');        /* consume ')' */
  }
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  return 0;                     /* return target identifier */
}  /* header() */

/*--------------------------------------------------------------------*/

static int trailer (SCANNER *scan)
{                               /* --- parse trailer */
  assert(scan);                 /* check the function argument */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* trailer() */

/*--------------------------------------------------------------------*/

REGRESS* reg_parse (SCANNER *scan, DIMID dim, int hdr)
{                               /* --- parse a regression polynomial */
  REGRESS *reg;                 /* parsed multivariate polynomial */

  assert(scan);                 /* check the function arguments */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (hdr && (header(scan) != 0))
    return NULL;                /* parse the header if necessary */
  reg = parse(scan, dim);       /* parse the regression function */
  if (!reg) return NULL;        /* (dimension, degree, parameters) */
  if (hdr && (trailer(scan) != 0)) {
    reg_delete(reg); return NULL; }
  return reg;                   /* return the parsed polynomial */
}  /* reg_parse() */

#endif  /* #ifdef REG_PARSE */
/*--------------------------------------------------------------------*/
#if 0

int reg_show (REGRESS *reg, FILE *file, int prec, double eps)
{                               /* --- describe a regression function */
  DIMID  i, k, n;               /* loop variables, counter */
  double *c;                    /* to traverse the coefficients */
  int    *e;                    /* table of exponents */
  char   *s;                    /* separators between terms */
  int    w;                     /* width of number output */

  assert(reg && file            /* check the function arguments */
  &&    (prec >= 0) && (eps >= 0));
  if (prec <= 0) prec = 6;      /* adapt the number of places */
  if (reg->cnt *prec <= 40) {   /* if the output fits into one line, */
    s = " ";      w = 0; }      /* print only a blank between terms, */
  else {                        /* if it does not fit into one line, */
    s = "\n    "; w = prec+7; } /* print one term per line */
  if (reg->max > reg->min) {    /* if logistic regression */
    fprintf(file, "z = (%g -%g) /(1 +exp(y))", reg->max, reg->min);
    fprintf(file, " +%g\n", reg->min);
  }                             /* print logistic transformation */
  fputs("y = ", file); n = 0;   /* start the function output */
  c = reg->cfs;                 /* traverse the reg. coefficients */
  if      (reg->dim <= 2) {     /* --- if single variable regression */
    for (i = reg->deg+1; --i > 1; c++) {
      if (fabs(*c) < eps) continue;   /* skip negligible terms */
      if (n++ > 0) fputs(s, file);    /* print a separator */
      fprintf(file, "%+*.*g x^%d", w, prec, *c, i);
    }                           /* print coeffs. of different powers */
    if (fabs(*c) >= eps) {      /* if linear term is not negligible */
      if (n++ > 0) fputs(s, file);
      fprintf(file, "%+*.*g x", w, prec, *c);
    } c++; }                    /* print the linear term */
  else if (reg->deg <= 1) {     /* --- if linear regression */
    for (i = 0; ++i < reg->dim; c++) {
      if (fabs(*c) < eps) continue;   /* skip negligible terms */
      if (n++ > 0) fputs(s, file);    /* print a separator */
      fprintf(file, "%+*.*g x_%d", w, prec, *c, i);
    } }                         /* print the linear terms */
  else if (reg->deg <= 2) {     /* --- if quadratic regression */
    for (i = 0; ++i < reg->dim; c++) {
      for (k = 0; ++k < i; c++) {
        if (fabs(*c) < eps) continue; /* skip negligible terms */
        if (n++ > 0) fputs(s, file);  /* print a separator */
        fprintf(file, "%+*.*g x_%d x_%d", w, prec, *c, k, i);
      }                         /* print the mixed terms */
      if (fabs(*c) < eps) continue;   /* skip negligible terms */
      if (n++ > 0) fputs(s, file);    /* print a separator */
      fprintf(file, "%+*.*g x_%d^2", w, prec, *c, i);
    }                           /* print the squares */
    for (i = 0; ++i < reg->dim; c++) {
      if (fabs(*c) < eps) continue;   /* skip negligible terms */
      if (n++ > 0) fputs(s, file);    /* print a separator */
      fprintf(file, "%+*.*g x_%d", w, prec, *c, i);
    } }                         /* print the linear terms */
  else {                        /* --- if general regression */
    e = reg->tab +(i = reg->dim-1);
    *--e = reg->deg +1;         /* init. the table of exponents to */
    while (--i > 0) *--e = 0;   /* x_n^(deg+1), i.e. "previous" state */
    for (i = reg->cnt; --i > 0; c++) {
      if (e[0] > 0) {           /* update the table of exponents: */
        e[1]++; e[0]--; }       /* if first exponent is not zero, */
      else {                    /* shift a 1 from first to second */
        for (k = 0; e[++k] <= 0; );   /* find first nonzero expo. */
        e[0] = e[k]-1; e[k] = 0; e[k+1]++;
      }                         /* move exponents back to first */
      if (fabs(*c) < eps) continue;   /* skip negligible terms */
      if (n++ > 0) fputs(s, file);    /* print a separator */
      fprintf(file, "%+*.*g", w, prec, *c);
      for (k = 0; ++k < reg->dim; ) {
        if      (e[k-1] == 1) fprintf(file, " x_%d", k);
        else if (e[k-1] >  1) fprintf(file, " x_%d^%d", k, e[k-1]);
      }                         /* print coefficients of */
    }                           /* the different products and */
  }                             /* a description the products */
  if (fabs(*c) >= eps) {        /* if constant term is not negligible */
    if (n++ > 0) fputs(s, file);/* print the constant term */
    fprintf(file, "%+*.*g", w, prec, *c); }
  else if (n <= 0)              /* if no term printed yet, */
    fputc('0', file);           /* print a zero as a fallback */
  fputc('\n', file);            /* terminate the output and */
  return ferror(file) ? -1 : 0; /* return the write status */
}  /* reg_show() */

#endif
